/*
** Module   : RecordIdentifier.java
** Abstract : Unique external identifier for a database record
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20070411 ADD   @32989 Created initial version. Unique external
**                               identifier for a database record.
** 002 ECF 20071128 CHG   @36053 Implement Serializable. Allows use over the
**                               network.
** 003 VMN 20130830              Made class not-final to TableIdentifier.
*/

package com.goldencode.p2j.persist;

import java.io.*;

/**
 * A combination of database table name and record ID (primary key), which
 * uniquely identifies a database record.
 */
public class RecordIdentifier
implements Serializable
{
   /** Table name */
   private final String table;
   
   /** Primary key of a database record */
   private final Serializable recordID;
   
   /**
    * Constructor.
    *
    * @param   table
    *          Table name.
    * @param   recordID
    *          Primary key.
    *
    * @throws  NullPointerException
    *          if any parameter is <code>null</code>.
    */
   public RecordIdentifier(String table, Serializable recordID)
   {
      if (table == null)
      {
         throw new NullPointerException("Table must be non-null");
      }
      if (recordID == null)
      {
         throw new NullPointerException("Record ID must be non-null");
      }
      
      this.table = table.intern();
      this.recordID = recordID;
   }
   
   
   /**
    * Get the database table name.
    *
    * @return  Table name.
    */
   public String getTable()
   {
      return table;
   }
   
   /**
    * Get the record's primary key.
    *
    * @return  Primary key.
    */
   public Serializable getRecordID()
   {
      return recordID;
   }
   /**
    * Overridden equivalence test which takes all data values of this
    * object into account.
    *
    * @param   o
    *          Another <code>LockRecord</code> instance to test for
    *          equivalence with this instance.
    *
    * @return  <code>true</code> if both instances have the same values
    *          for table and record ID, else <code>false</code>.
    */
   public boolean equals(Object o)
   {
      if (!(o instanceof RecordIdentifier))
      {
         return false;
      }
      
      RecordIdentifier that = (RecordIdentifier) o;
      
      return (table.equals(that.table) &&
              recordID.equals(that.recordID));
   }
   
   /**
    * Overridden implementation to maintain consistency with overridden
    * {@link #equals} method.
    *
    * @return  Hash code for this instance.
    */
   public int hashCode()
   {
      int result = 17;
      result = 37 * result + table.hashCode();
      result = 37 * result + recordID.hashCode();
      
      return result;
   }
   
   /**
    * Compose a string representation of this object for assistance in
    * debugging and testing.
    *
    * @return  String representation of this object's internal state.
    */
   public String toString()
   {
      StringBuffer buf = new StringBuffer();
      
      buf.append(table);
      buf.append(':');
      buf.append(recordID);
      
      return buf.toString();
   }
}
