/*
** Module   : LockManagerMultiplexer.java
** Abstract : Defines a remote API for database record lock management
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20071018 ADD  @35552  Created initial version. Defines a remote API
**                               for database record lock management.
** 002 ECF 20071128 CHG  @36054  Replace LockType with int for method return
**                               values and parameters. This prevents us from
**                               having to support LockType as Serializable
**                               objects, which is desirable for several
**                               reasons:
**                               1) less data to send over the network;
**                               2) allows us to avoid creating multiple
**                                  instances of each LockType type, avoiding
**                                  a lot of unnecessary working set.
** 003 VMN 20130903              Added lockTable method.
*/

package com.goldencode.p2j.persist.remote;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * Interface which defines a client-server API for remote database record lock
 * management.  {@link RemoteLockManager} invokes the methods of this
 * interface on a local proxy for a remote object, in order to access the lock
 * management services of the remote P2J server which is authoritative for a
 * particular database.
 * <p>
 * Since the remote P2J server may manage multiple databases, but only a
 * single instance of an implementor of this interface will be available at
 * any given remote server, lock management service requests must be
 * multiplexed by database.  For this reason, each of the methods defined by
 * this interface accepts an integer ID, which uniquely identifies the target
 * database at the remote server.
 * <p>
 * The database ID to use for a particular physical database is determined by
 * calling {@link RemoteMultiplexer#getMultiplexID(Object)}.  This is the
 * responsibility of the <code>RemoteLockManager</code> and must be done
 * before any lock management service is requested.
 * <p>
 * This interface effectively mirrors that of {@link
 * com.goldencode.p2j.persist.LockManager the LockManager API}, except for the
 * additional database multiplex identifier which is the first parameter to
 * each method.
 * 
 * @author ECF
 */
public interface LockManagerMultiplexer
extends RemoteMultiplexer<String>
{
   /**
    * Attempt to obtain the specified lock type on a particular record.  If
    * the lock cannot be obtained immediately, the current thread will block
    * until the lock becomes available and is obtained, unless the requested
    * lock type is a "no-wait" lock.  In the latter case, an exception is
    * raised, indicating that the lock is not available.  The current thread
    * continues once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying
    * <code>LockType.NONE</code> for the <code>lockType</code> parameter.
    * <p>
    * The normal return of this method indicates that the lock request (or
    * lack thereof) completed successfully, and that the current context now
    * holds a lock of the requested type.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   lockType
    *          Type of lock to be obtained.  <code>LockType.NONE</code> is
    *          used to release an existing lock, and to continue with no
    *          lock.
    * @param   ident
    *          ID which uniquely identifies the record being locked.
    * @param   update
    *          <code>true</code> if the lock state for the current record
    *          should be updated;  else <code>false</code>.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    */
   public void lock(int database,
                    int lockType,
                    RecordIdentifier ident,
                    boolean update)
   throws LockUnavailableException;

   /**
    * Attempt to obtain the specified lock type on a table. If the lock cannot be obtained
    * immediately, the current thread will block until the lock becomes available and is
    * obtained, unless the requested lock type is a "no-wait" lock.  In the latter case,
    * an exception is raised, indicating that the lock is not available. The current thread
    * continues once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying
    * <code>LockType.NONE</code> for the <code>lockType</code> parameter.
    * <p>
    * Method returns previous lock type (so that calling code knows what to set it back to after
    * bulk delete).
    * The normal return of this method indicates that the lock request (or lack thereof)
    * completed successfully, and that the current context now holds a lock of the requested
    * type.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the remote server.
    * @param   lockType
    *          Type of lock to be obtained.  <code>LockType.NONE</code> is used to release
    *          an existing lock, and to continue with no lock.
    * @param   table
    *          Table name.
    * @param   update
    *          <code>true</code> if the lock state for the current record should be updated;
    *          else <code>false</code>.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    *
    * @return  previous lock type.
    */
   public LockType lockTable(int database,
                             int lockType,
                             String table,
                             boolean update)
   throws LockUnavailableException;

   /**
    * Query the lock type currently held by the current context for the
    * specified database record.
    * <p>
    * This method should always return the <b>non-</b><code>NO_WAIT</code>
    * variants of a lock type.  That is, even if the actual lock type is
    * <code>LockType.EXCLUSIVE_NO_WAIT</code> or
    * <code>LockType.SHARE_NO_WAIT</code>, the simpler types of
    * <code>LockType.EXCLUSIVE</code> and <code>LockType.SHARE</code>,
    * respectively, should be returned.  The lack of a lock is required to
    * return <code>LockType.NONE</code> rather than <code>null</code>.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   ident
    *          ID which uniquely identifies the record being queried.
    *
    * @return  The lock type currently held by the current context, never
    *          <code>null</code>.
    */
   public int queryLock(int database, RecordIdentifier ident);
}
